#!/usr/bin/env python
import time
import traceback
import logging

from PyQt5 import QtWidgets, QtCore
from PyQt5.QtWidgets import QApplication, QMainWindow
from pyqtgraph import _connectCleanup as pyqtgraph_connectCleanup

from nspyre.config import load_meta_config
from nspyre.definitions import CLIENT_META_CONFIG_PATH
from nspyre.inserv import InservGateway
from nspyre.spyrelet import SpyreletLauncher, load_all_spyrelets

__package__ = 'nspyre.gui'

from .widgets.param_widget import ParamWidget
from .widgets.save_widget import Save_Widget

logger = logging.getLogger(__name__)


class ProgressBar(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.vars = []
        # self.iterators = []
        layout = QtWidgets.QVBoxLayout()
        self.pbar = QtWidgets.QProgressBar()
        self.pbar.setTextVisible(True)
        self.text = QtWidgets.QLabel()
        layout.addWidget(self.pbar)
        layout.addWidget(self.text)
        self.setLayout(layout)
    
    def reset(self):
        self.vars = []
        # self.iterators = []
        self.text.clear()

    def call_iter(self, iterable, max_val):
        # self.iterators.append(iter(iterable))
        self.vars.append({'val': 0, 'max': max_val, 'start': time.time(), 'last': time.time(), 'avg': 0, 'tot': 0, 'rem': '?', 'per': 0})
        self.text.setText('started')

        if max_val != '?':
            self.pbar.setValue(0)
            self.pbar.setRange(0, max_val)

    def call_next(self):
        self.vars[-1]['val'] += 1
        t = time.time()
        self.vars[-1]['avg'] = t-self.vars[-1]['last']
        self.vars[-1]['last'] = t
        self.vars[-1]['tot'] = t-self.vars[-1]['start']

        if self.vars[-1]['max'] != '?':
            self.pbar.setValue(self.vars[-1]['val'])
            self.pbar.setRange(0, self.vars[-1]['max'])
            self.vars[-1]['per'] = 100*self.vars[-1]['val']/self.vars[-1]['max']
            self.vars[-1]['rem'] = (self.vars[-1]['max'] - self.vars[-1]['val'])*self.vars[-1]['avg']
        s_with_max = '[{per:.0f}% {val:.0f}/{max:.0f} [{tot:.0f}s<{rem:.0f}s] {avg:.2f}s/it]'
        s_without_max = '[{val:.0f}[{tot:.0f}s] {avg:.2f}s/it]'
        self.text.setText('\t'.join([(s_without_max if d['max'] == '?' else s_with_max).format(**d)for d in self.vars]))
        QtWidgets.QApplication.processEvents()
    
    def call_stopiter(self):
        self.vars = self.vars[:-1]
        if not self.vars == [] and self.vars[-1]['max'] != '?':
            self.pbar.setValue(self.vars[-1]['max'])
            self.pbar.setRange(0, self.vars[-1]['max'])
            QtWidgets.QApplication.processEvents()
        # self.iterators = self.iterators[:-1]


class SpyreletLauncherWidget(QtWidgets.QWidget):
    def __init__(self, spyrelet, parent=None):
        self.spyrelet = spyrelet
        self.progress_bar = ProgressBar()
        self.launcher = SpyreletLauncher(spyrelet)
        self.param_w = ParamWidget(self.launcher.params)
        self.param_w.set(**self.launcher.get_defaults())
        super().__init__(parent=parent)

        # Build ctrl pannel
        ctrl_pannel = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout()
        self.run_btn = QtWidgets.QPushButton('Run')
        self.set_defaults_btn = QtWidgets.QPushButton('Set Defaults')
        self.save_btn = QtWidgets.QPushButton('Save')
        layout.addWidget(self.run_btn)
        layout.addWidget(self.save_btn)
        ctrl_pannel.setLayout(layout)

        # Build main layout
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(ctrl_pannel)
        layout.addWidget(self.progress_bar)
        layout.addWidget(self.param_w)
        layout.addStretch()
        self.setLayout(layout)

        # Connect signal
        self.run_btn.clicked.connect(self.run)
        self.set_defaults_btn.clicked.connect(self.set_defaults)
        self.save_btn.clicked.connect(self.save_data)

        self.run_thread = None

    def run(self):
        if self.run_thread is None or self.run_thread.isFinished():
            self.progress_bar.reset()
            param_dict = self.param_w.get()
            self.run_thread = SpyreletRunThread(self.launcher, param_dict)
            self.run_thread.start()
            self.run_btn.setText('Stop')
            self.run_thread.finished.connect(lambda: self.run_btn.setText('Run'))
            self.run_thread.progressed_iter.connect(self.progress_bar.call_iter)
            self.run_thread.progressed_next.connect(self.progress_bar.call_next)
            self.run_thread.progressed_stopiter.connect(self.progress_bar.call_stopiter)
        else:
            self.run_thread.stop_requested.emit()

    def set_defaults(self):
        self.spyrelet.set_defaults(**self.param_w.get())

    def save_data(self):
        self.save_w = Save_Widget(self.spyrelet)
        self.save_w.show()

class SpyreletRunThread(QtCore.QThread):
    """Qt Thread which monitors for changes to qather a collection or a database
    and emits a signal when something happens.
    """
    progressed_iter = QtCore.pyqtSignal(object, object)  # This will be emitted when iter is called on progress is called in the spyrelet
    progressed_next = QtCore.pyqtSignal()  # This will be emitted when next is called on progress is called in the spyrelet
    progressed_stopiter = QtCore.pyqtSignal()  # This will be emitted when next is called on progress is called in the spyrelet
    stop_requested = QtCore.pyqtSignal()  # This will be emitted externally to stop the execution of a spyrelet prematurly

    def __init__(self, launcher, param_dict):
        super().__init__()
        self.param_dict = param_dict
        self.launcher = launcher
        self.spyrelet = self.launcher.spyrelet

        class ProgressIter():
            def __init__(_self, iterable):
                _self.iterable = iterable

            def __iter__(_self):
                try:
                    max_val = len(_self.iterable)
                except TypeError:
                    traceback.print_exc()
                    max_val = '?'
                _self.iterable = iter(_self.iterable)
                self.progressed_iter.emit(_self.iterable, max_val)
                return _self

            def __next__(_self):
                try:
                    val = next(_self.iterable)
                    self.progressed_next.emit()
                except StopIteration as e:
                    self.progressed_stopiter.emit()
                    raise e
                return val

        self.progress = ProgressIter
        self.stop_requested.connect(self.stop_run)

    def progress(self, iterator):
        for x in iterator:
            self.progressed.emit(iterator)
            yield x

    def run(self):
        self.launcher.run(progress=self.progress, **self.param_dict)

    def stop_run(self):
        self.spyrelet.stop()


class CombinedSpyreletWindow(QMainWindow):
    def __init__(self, gateway, spyrelets=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Set the main window layout to consist of vertical boxes.
        # The QVBoxLayout class lines up widgets vertically.
        layout = QtWidgets.QVBoxLayout()

        self.selector = QtWidgets.QComboBox()
        container = QtWidgets.QWidget()
        layout.addWidget(self.selector)
        layout.addWidget(container)

        #Create the launchers
        spyrelets = load_all_spyrelets(gateway) if spyrelets is None else spyrelets
        self.launchers = {name: SpyreletLauncherWidget(s) for name, s in spyrelets.items()}

        #Add to layout
        stacked_layout = QtWidgets.QStackedLayout()
        names = list(self.launchers.keys())
        names.sort()
        for n in names:
            stacked_layout.addWidget(self.launchers[n])
        container.setLayout(stacked_layout)
        self.setWindowTitle('NSpyre Spyrelet Window: {}'.format(names[0]))

        self.selector.addItems(names)
        self.container_layout = stacked_layout
        self.selector.currentTextChanged.connect(self.change_widget)
        w = QtWidgets.QWidget()
        w.setLayout(layout)
        self.setCentralWidget(w)
        self.show()

    def change_widget(self, name):
        self.container_layout.setCurrentWidget(self.launchers[name])
        self.setWindowTitle('NSpyre Spyrelet Window: {}'.format(name))


if __name__ == '__main__':
    import logging
    import sys
    from PyQt5.QtCore import Qt
    from nspyre.gui import NSpyreApp
    from nspyre.misc import nspyre_init_logger

    nspyre_init_logger(logging.INFO)

    logger.info('starting Spyrelets...')
    if hasattr(Qt, 'AA_EnableHighDpiScaling'):
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
        QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
    app = NSpyreApp([sys.argv])
    pyqtgraph_connectCleanup()
    config_path = load_meta_config(CLIENT_META_CONFIG_PATH)
    with InservGateway(config_path) as isg:
        combined_spyrelet_window = CombinedSpyreletWindow(isg)
        sys.exit(app.exec())



from PyQt5 import QtWidgets
import pyqtgraph as pg
from collections import OrderedDict

from nspyre.definitions import Q_
from nspyre.misc import RangeDict

from .spinbox import SpinBox


class Rangespace(QtWidgets.QWidget):
    def __init__(self, units=None, parent=None, default=None):
        super().__init__(parent=parent)
        if units is None:
            units = 'dimensionless'
        self.units = units
        self.params_layout = QtWidgets.QStackedLayout()
        self.generate_range_widgets(units=units)
        self.range_func_combo = pg.ComboBox(items=list(self.range_widgets.keys()))
        self.range_func_combo.currentIndexChanged[str].connect(self.set_widget)

        layout = QtWidgets.QFormLayout()
        layout.addRow('Function', self.range_func_combo)
        select = QtWidgets.QWidget()
        select.setLayout(layout)
        params = QtWidgets.QWidget()
        params.setLayout(self.params_layout)
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(select)
        layout.addWidget(params)
        self.setLayout(layout)
        if not default is None:
            self.set(default)

    def get(self):
        func = self.range_func_combo.value()
        params = self.range_widgets[func].get()
        params['func'] = func
        return RangeDict(**params)

    def set(self, rangedict):
        d = RangeDict(**rangedict)
        func = d.pop('func')
        self.range_func_combo.setValue(func)
        fields = self.range_widgets[func].setter_methods.keys()
        for name in d:
            if not name in fields:
                d.pop(name)
        self.range_widgets[func].set(**d)

    def set_widget(self, name):
        self.params_layout.setCurrentWidget(self.range_widgets[name])

    def generate_range_widgets(self, units=None):
        self.range_widgets = OrderedDict()
        
        def add(name, params):
            self.range_widgets[name] = ParamWidget(OrderedDict(params))
            self.params_layout.addWidget(self.range_widgets[name])

        f_param = lambda default:   {
                                    'type': float,
                                    'units': units,
                                    'default':default,
                                    }
        i_param_dimless = lambda default:   {
                                                'type': int,
                                                'default':default,
                                                'positive': True,
                                            }
        i_param = lambda default:   {
                                        'type': int,
                                        'default':default,
                                        'units':units,
                                    }

        default_val = 1 if units is None else Q_(1, units).to_base_units().m

        params = [
            ('start', f_param(0)),
            ('stop',  f_param(default_val)),
            ('num', i_param_dimless(10)),
            ('endpoint', {'type':bool, 'default':True})
        ]
        add('linspace', params)

        params = [
            ('start', i_param(0)),
            ('stop', i_param(10*int(default_val))),
            ('step', i_param(1))
        ]
        add('arange', params)

        params = [
            ('start', f_param(default_val)),
            ('stop', f_param(10*default_val)),
            ('num', i_param_dimless(10))
        ]
        add('logspace', params)


class ParamWidget(QtWidgets.QWidget):

    def __init__(self, parameters, parent=None):
        super().__init__(parent=parent)
        self.parameters = parameters
        self.init_ui()
        return

    def init_ui(self):
        layout = QtWidgets.QFormLayout()
        getter_methods = dict()
        setter_methods = dict()
        for parameter_name, parameter_opts in self.parameters.items():
            typ = parameter_opts.get('type', str)
            default = parameter_opts.get('default')
            if typ == str:
                w = QtWidgets.QLineEdit(default)
                activation_method = w.text
                getter_method = w.text
                setter_method = w.setText
            elif typ == int or typ == float:
                default_opts = {
                    # 'suffix': parameter_opts.get('units', ''),
                    # 'siPrefix': False,
                    'dec': True,
                    'int': typ == int,
                    'minStep': 1 if typ == int else 0.1,
                    'step': 1 if typ == int else 0.1,
                    'bounds': [None, None],
                    'decimals': 8,
                }
                opts = dict()
                if parameter_opts.get('nonnegative', False):
                    opts['bounds'] = [0, None]
                if typ == int and parameter_opts.get('positive', False):
                    opts['bounds'] = [1, None]
                for k in default_opts.keys():
                    opts[k] = parameter_opts.get(k, default_opts[k])

                w = SpinBox(unit=parameter_opts.get('units'), **opts)
                # w = pg.SpinBox(**opts)
                w.setValue(default)
                # getter_method = w.value
                getter_method = w.unit_value
                setter_method = w.setValue
            elif typ == bool:
                w = QtWidgets.QCheckBox()
                getter_method = w.isChecked
                setter_method = w.setChecked
                default = default if default is not None else False
                w.setChecked(default)
            elif typ == object:
                w = parameter_opts['widget']
                getter_method = parameter_opts['getter']
                setter_method = parameter_opts.get('setter', None)
            elif typ == range:
                w = Rangespace(units=parameter_opts.get('units'))
                getter_method = w.get
                setter_method = w.set
                if not default is None:
                    w.set(default)
            elif typ == list:
                w = pg.ComboBox()
                items = parameter_opts.get('items')
                if type(items) == list:
                    w.setItems(items)
                else:
                    raise ValueError('items must be specified and of type list to add a list type entry to ParamWidget')
                getter_method = w.value
                setter_method = w.setText
                if not default is None:
                    try:
                        w.setText(default)
                    except ValueError:
                        pass
            elif typ == dict:
                w = pg.ComboBox()
                items = parameter_opts.get('items')
                if type(items) == dict:
                    w.setItems(items)
                else:
                    raise ValueError('items must be specified and of type dict to add a dict type entry to ParamWidget')
                getter_method = w.value
                setter_method = w.setValue
                if not default is None:
                    w.setValue(default)

            layout.addRow(parameter_name, w)
            getter_methods[parameter_name] = getter_method
            setter_methods[parameter_name] = setter_method
        self.getter_methods = getter_methods
        self.setter_methods = setter_methods
        self.setLayout(layout)
        return

    def get(self):
        return {name: method() for name, method in self.getter_methods.items()}

    def set(self, **kwargs):
        for k, v in kwargs.items():
            setter = self.setter_methods[k]
            if setter is not None:
                setter(v)
        return

    def save_state(self):
        state = self.get()
        return state

    def load_state(self, state):
        self.set(**state)



import datetime
import logging
import sys

from pathlib import Path


class StreamToLog(object):
    """Fake stream object that redirects writes to a logger"""
    def __init__(self, logger, log_level, terminator):
        self.logger = logger
        self.log_level = log_level
        self.write_buffer = ''
        self.terminator = terminator

    def write(self, buff):
        """Override the stream write method"""
        while buff:
            # if buff contains a terminator, we should split into multiple
            # log messages, separated by the terminator
            if self.terminator in buff:
                before,_,after = buff.partition(self.terminator)
                self.write_buffer += before
                self.logger.log(self.log_level, self.write_buffer)
                self.write_buffer = ''
                buff = after
            # otherwise, just add it to the buffer for writing later when
            # a terminator is received
            else:
                self.write_buffer += buff
                buff = ''

    def flush(self):
        if self.write_buffer:
            self.write(self.terminator)


def nspyre_init_logger(log_level, log_path=None, log_path_level=None, prefix=None, file_size=None):
    """Initialize system-wide logging to stdout/err and, optionally, a file
    log_level: log messages of lower severity than this will not be sent to stdout/err (e.g. logging.INFO)
    log_path: if a file, log to that file; if a directory, generate a log file 
                name and create a new log file in that directory; if None, only log to stdout/err
    log_path_level: logging level for the log file - leave as None for same as log_level
    prefix: if a directory was specified for log_path, prepend this string
                to the log file name
    file_size: maximum log file size; if this size is exceeded, the log file is rotated according
                to RotatingFileHandler - https://docs.python.org/3/library/logging.handlers.html"""

    root_logger = logging.getLogger()
    # the root logger will accept all messages
    root_logger.setLevel(logging.DEBUG)

    # log format for stdout messages
    stdout_formatter = logging.Formatter('%(asctime)s.%(msecs)03d [%(levelname)8s] %(message)s', '%Y-%m-%d %H:%M:%S')
    # create stdout log handler
    stdout_handler = logging.StreamHandler(stream=sys.stdout)
    stdout_handler.setLevel(log_level)
    stdout_handler.setFormatter(stdout_formatter)
    root_logger.addHandler(stdout_handler)

    # all stderr messages will now be redirected to a special logger
    stderr_logger = logging.getLogger('stderr')
    stderr_logger.propagate = False
    sys.stderr = StreamToLog(stderr_logger, logging.CRITICAL, '\n')

    # log format for stderr messages
    stderr_formatter = logging.Formatter('[stderr] %(message)s')
    # create stderr log handler
    stderr_handler = logging.StreamHandler(stream=sys.stdout)
    stderr_handler.setLevel(log_level)
    stderr_handler.setFormatter(stderr_formatter)
    stderr_logger.addHandler(stderr_handler)

    # if a log file / folder was specified
    if log_path:
        # resolve relative paths
        if not log_path.is_absolute():
            log_path = Path.cwd() / log_path
        log_path = log_path.resolve()

        if log_path.is_file():
            pass
        elif log_path.is_dir():
            # log to a file in the folder
            file_name = datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S.log'.format())
            # prepend the prefix if present
            if prefix:
                file_name = '{}-{}'.format(prefix, file_name)
            log_path = log_path / Path(file_name)

        # create the file handler
        if file_size:
            file_handler = logging.handlers.RotatingFileHandler(log_path, maxBytes=file_size, backupCount=1000)
        else:
            file_handler = logging.FileHandler(log_path)
        file_formatter = logging.Formatter('[stderr] %(asctime)s.%(msecs)03d [%(levelname)s] (%(filename)s:%(lineno)s) %(message)s', '%Y-%m-%d %H:%M:%S')
        file_handler.setFormatter(file_formatter)
        if log_path_level:
            file_handler.setLevel(log_path_level)
        else:
            file_handler.setLevel(log_level)
        root_logger.addHandler(file_handler)
        stderr_logger.addHandler(file_handler)



# from nspyre.definitions import Q_
# from nspyre.gui import Plot1D, Plot2D, PlotFormatInit, PlotFormatUpdate
# from nspyre.gui.data_handling import load_data
# from nspyre.spyrelet import Spyrelet, load_all_spyrelets, unload_all_spyrelets
# from nspyre.widgets import LinePlotWidget, HeatmapPlotWidget
# from nspyre.inserv import InservGateway

# __all__ = ['Plot1D', 'Plot2D', 'PlotFormatInit', 'PlotFormatUpdate', 
#            'Spyrelet', 'LinePlotWidget', 'HeatmapPlotWidget', 'Q_', 
#            'load_data', 'instrument_manager', 'load_all_spyrelets', 
#            'unload_all_spyrelets']

__version__ = '0.4.9'



###########################
# imports
###########################

# std
import logging

# nspyre
from nspyre.spyrelet.spyrelet import Spyrelet
from nspyre.definitions import Q_

# from nspyre.gui import LinePlotWidget, Plot1D, Plot2D, PlotFormatInit, PlotFormatUpdate

###########################
# classes
###########################

class SinglePoint(Spyrelet):
    REQUIRED_DEVICES = [
        'fake_sg',
        'fake_daq',
        'fake_pellicle',
    ]

    PARAMS = {
        'amp': {
            'type': float,
            'units':'V'},
        'freq': {
            'type': float,
            'units':'GHz'},
    }

    def main(self, freq, amp):
        # open the pellicle
        self.fake_pellicle.opened = True
        # set the frequency / amplitude of the sg
        self.fake_sg.amplitude = amp
        self.fake_sg.frequency = freq
        # make an analog input read
        val = self.fake_daq.ain[0]
        # gather the data
        self.acquire({
            'freq': freq,
            'A': amp,
            'result': val,
        })

        # close the pellicle
        self.fake_pellicle.opened = False

    def initialize(self, freq, amp):
        logging.info('initializing [{}]...'.format(self.name))

    def finalize(self, freq, amp):
        logging.info('finalizing [{}]...'.format(self.name))

class FreqAmpSweep(Spyrelet):
    REQUIRED_DEVICES = [
        'fake_sg',
        'fake_daq',
        'fake_pellicle',
        'lantz_scope'
    ]

    PARAMS = {
        'amp_range': {
            'type': range,
            'units':'V'},
        'freq_range': {
            'type': range,
            'units':'GHz'},
    }

    def main(self, freq_range, amp_range):
        # flip up the pellicle
        self.fake_pellicle.opened = True

        # iterate over ampitudes
        for a, amp in enumerate(self.progress(amp_range)):
            self.fake_sg.amplitude = amp
            # iterate over frequencies
            for f, freq in enumerate(self.progress(freq_range)):
                self.fake_sg.frequency = f
                # measure an analog input voltage
                val = self.fake_daq.ain[0]
                # gather the data
                self.acquire({
                    'ind': i,
                    'f': f,
                    'a': a,
                    'A': amplitude,
                    'result': val,
                })

        rand_array = self.lantz_scope.measure()
        self.fake_pellicle.opened = False

    def initialize(self, freq_range, amp_range):
        logging.info('initializing [{}]...'.format(self.name))

    def finalize(self, freq_range, amp_range):
        logging.info('finalizing [{}]...'.format(self.name))

    # @Plot1D
    # def plot_results(df, cache):
    #     return {'result':[df.ind.values, df.result.values]}

    # @Plot1D
    # def plot_f(df, cache):
    #     return {'f':[df.ind.values, df.f.values], 'f2':[df.ind.values, 2*df.f.values]}

    # @PlotFormatInit(LinePlotWidget, ['plot_f', 'plot_results'])
    # def init_formatter(plot):
    #     plot.xlabel = 'My x axis (a.u.)'
    #     plot.ylabel = 'Signal (in bananas)'



"""
This file tests the functionality of the instrument server and gateway. It 
presumes that mongodb and the instrument server have been started from 
conftest.py by pytest.

Author: Jacob Feder
Date: 11/12/2020
"""

###########################
# imports
###########################

# std
import logging

# nspyre
from nspyre.inserv.gateway import InservGateway
from nspyre.definitions import Q_

###########################
# globals
###########################

logger = logging.getLogger(__name__)

###########################
# tests
###########################

class TestInserv:
    def test_feats_get_set(self, gateway):
        """test basic feat get/set"""
        gateway.tserv.fake_sg.amplitude = Q_(1.0, 'V')
        assert gateway.tserv.fake_sg.amplitude == Q_(1.0, 'V')
        gateway.tserv.fake_sg.amplitude = Q_(10.0, 'V')
        assert gateway.tserv.fake_sg.amplitude == Q_(10.0, 'V')

    def test_feats_units(self, gateway):
        """test get/set with different pint units"""
        gateway.tserv.fake_sg.amplitude = Q_(0.1, 'V')
        assert gateway.tserv.fake_sg.amplitude == Q_(100.0, 'mV')
        gateway.tserv.fake_sg.amplitude = Q_(10, 'mV')
        assert gateway.tserv.fake_sg.amplitude == Q_(0.01, 'V')

    def test_dictfeats_get_set(self, gateway):
        """test basic dictfeat get/set"""
        for k in range(1, 10):
            gateway.tserv.fake_daq.dout[k] = False
            assert gateway.tserv.fake_daq.dout[k] == False
        for k in range(1, 10):
            gateway.tserv.fake_daq.dout[k] = True
            assert gateway.tserv.fake_daq.dout[k] == True

    def test_dictfeats_ro(self, gateway):
        """test read-only dictfeats"""
        gateway.tserv.fake_daq.reset_din(False)
        for k in range(1, 10):
            assert gateway.tserv.fake_daq.din[k] == False
            gateway.tserv.fake_daq.toggle_din(k)
            assert gateway.tserv.fake_daq.din[k] == True



"""
This module manages and centralizes connections to one or more instrument
servers. All instrument server connections should be done through the
InservGateway class.

Author: Jacob Feder
Date: 7/11/2020
"""
import os
import logging

import rpyc

from nspyre.config import get_config_param, load_config, load_meta_config
from nspyre.definitions import CLIENT_META_CONFIG_PATH, Q_, RPYC_CONN_TIMEOUT, RPYC_SYNC_TIMEOUT
from nspyre.errors import InservGatewayError
from nspyre.misc import register_quantity_brining

# for properly serializing/deserializing quantity objects using the local
# pint unit registry
register_quantity_brining(Q_)

logger = logging.getLogger(__name__)

CONFIG_GATEWAY_SETTINGS = 'instrument_servers'
CONFIG_GATEWAY_DEVICES = 'devices'

# Temporary monkey patching of rpyc to implement synchronous about_to_disconnect feature
# Need to define consts.HANDLE_ABOUT_TO_CLOSE before consts is import by protocol (this is done
# in the load of nspyre.inserv.inserv)
from .inserv import (InstrumentConnection as nspyre_InstrumentConnection,
                     InstrumentService as nspyre_InstrumentService,
                     VoidInstrumentService as nspyre_VoidInstrumentService)
from rpyc.core.protocol import consts

# Need to monkey patch VoidService for rpyc.utils.factory.connect_stream
from rpyc.core.service import Service
Service._protocol = nspyre_InstrumentConnection
Service = nspyre_InstrumentService
from rpyc.core.service import VoidService
VoidService = nspyre_VoidInstrumentService


class InservGateway:
    """Loads a configuration file, then attempts to connect to all 
    instrument servers

    This class is a representation of the instrument server from the
    client's perspective. It contains server connection information, and an
    instance variable object for each device connected to the remote instrument
    server
    """
    def __init__(self, config_file=None):
        # if the config file isn't specified, get it from the meta-config
        if not config_file:
            config_file = load_meta_config(CLIENT_META_CONFIG_PATH)
        # config dictionary
        self.config = {}
        # dictionary of available rpyc instrument servers
        # key is the server string id, value is a tuple (rpyc conn, bg thread)
        # e.g. {'local1': (rpyc.core.protocol.Connection, 
        #                  rpyc.utils.helpers.BgServingThread),
        #       'remote1': ...}
        self._servers = {}
        self.config = None
        self.reload_config(config_file)
        self.reconnect_servers()

    def reconnect_servers(self):
        """Attempt connection to all of the instrument servers specified in the config"""
        servers, _ = get_config_param(self.config, [CONFIG_GATEWAY_SETTINGS])
        # iterate through servers
        for server_name in servers:
            # only try connecting if there isn't already a connection
            if server_name not in self._servers:
                ip, _ = get_config_param(self.config, [CONFIG_GATEWAY_SETTINGS, server_name, 'ip'])
                port, _ = get_config_param(self.config, [CONFIG_GATEWAY_SETTINGS, server_name, 'port'])
                try:
                    self.connect_server(server_name, ip, port)
                except InservGatewayError:
                    logger.error('Couldn\'t connect to instrument server [{}]'.format(server_name))

    def disconnect_servers(self):
        """Attempt disconnection from all of the instrument servers"""
        for s in list(self._servers):
            self.disconnect_server(s)

    def connect_server(self, s_id, s_addr, s_port):
        """Attempt connection to an instrument server"""
        try:
            # connect to the rpyc server running on the instrument server
            # and start up a background thread to fullfill requests on the
            # client side
            conn = rpyc.connect(s_addr, s_port,
                                config={'allow_pickle': True,
                                        'timeout': RPYC_CONN_TIMEOUT,
                                        'sync_request_timeout': RPYC_SYNC_TIMEOUT})

            bg_serving_thread = rpyc.BgServingThread(conn)
            
            # this allows the instrument server to have full access to this
            # client's object dictionaries - appears necessary for lantz
            conn._config['allow_all_attrs'] = True

            self._servers[s_id] = (conn, bg_serving_thread)
        except BaseException:
            raise InservGatewayError('Failed to connect to instrument server [{}] at address [{}]'.format(s_id, s_addr)) from None
        logger.info('instrument server gateway connected to instrument server [{}]'.format(s_id))

    def disconnect_server(self, s_id):
        """Disconnect from an instrument server and remove it's associated 
        devices"""
        try:
            conn = self._servers[s_id][0]
            bg_serving_thread = self._servers[s_id][1]
            bg_serving_thread.stop()
            conn.close()
            del self._servers[s_id]
        except BaseException:
            raise InservGatewayError('Failed to disconnect from instrument server [{}]'.format(s_id)) from None
        logger.info('instrument server gateway disconnected from server [{}]'.format(s_id))

    def servers(self):
        """Return a dictionary containing 'server name' mapped to
        an rpyc conn object"""
        servers_dict = {}
        for s in self._servers:
            servers_dict[s] = self._servers[s][0]

        return servers_dict

    def __getattr__(self, attr):
        """Allow the user to access the server objects directly using
        e.g. gateway.server1.sig_gen.frequency notation"""
        if attr in self._servers:
            return self._servers[attr][0].root
        else:
            raise AttributeError('\'{}\' object has no attribute \'{}\''.format(self.__class__.__name__, attr))

    def reload_config(self, filename):
        """Reload the config file"""
        self.config = load_config(filename)

    def __enter__(self):
        """Python context manager setup"""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Python context manager teardown"""
        self.disconnect_servers()



"""The class defining the NSpyre Instrument Server RPyC Service.

This module:
    - loads the server config file
    - connects to all of the instruments specified in the config files
    - creates a RPyC (python remote procedure call) server to allow remote
        machines to access the instruments (should be done using InservGateway)

Copyright (c) 2020, Alexandre Bourassa, Michael Solomon, Jacob Feder
All rights reserved.

This work is licensed under the terms of the 3-Clause BSD license.
For a copy, see <https://opensource.org/licenses/BSD-3-Clause>.
"""
from pathlib import Path
import logging
import threading
import time
import copy
import functools

import pymongo
import rpyc
from rpyc.utils.server import ThreadedServer

from lantz import Q_, DictFeat
from pimpmyclass.helpers import DictPropertyNameKey
from pint import Quantity

from nspyre.config import get_config_param, load_config, load_meta_config
from nspyre.definitions import CONFIG_MONGO_ADDR_KEY, MONGO_CONNECT_TIMEOUT, MONGO_RS, MONGO_SERVERS_KEY, RPYC_SYNC_TIMEOUT, SERVER_META_CONFIG_PATH
from nspyre.errors import EntryNotFoundError, InstrumentServerError
from nspyre.misc import load_class_from_file, load_class_from_str, register_quantity_brining

# for properly serializing/deserializing quantity objects using the local
# pint unit registry
register_quantity_brining(Q_)

logger = logging.getLogger(__name__)

CONFIG_SERVER_SETTINGS = 'server_settings'
CONFIG_SERVER_DEVICES = 'devices'
CONFIG_SERVER_DEVICE_LANTZ_CLASS = 'lantz_class'
CONFIG_SERVER_DEVICE_CLASS_FILE = 'class_file'
CONFIG_SERVER_DEVICE_CLASS_NAME = 'class'

RPYC_SERVER_STOP_EVENT = threading.Event()



rpyc.core.consts.HANDLE_ABOUT_TO_CLOSE = 21
from rpyc.core.protocol import consts


class InstrumentConnection(rpyc.Connection):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _pre_cleanup(self):  # IO
        self._local_root.on_about_to_disconnect(self)

    def close(self, _catchall=True):  # IO
        """closes the connection, releasing all held resources"""
        if self._closed:
            return
        self._closed = True
        try:
            self.sync_request(consts.HANDLE_ABOUT_TO_CLOSE)
            self._async_request(consts.HANDLE_CLOSE)
        except EOFError:
            pass
        except Exception:
            if not _catchall:
                raise
        finally:
            self._cleanup(_anyway=True)

    @classmethod
    def _request_handlers(cls):  # request handlers
        return {
            consts.HANDLE_PING: cls._handle_ping,
            consts.HANDLE_CLOSE: cls._handle_close,
            consts.HANDLE_GETROOT: cls._handle_getroot,
            consts.HANDLE_GETATTR: cls._handle_getattr,
            consts.HANDLE_DELATTR: cls._handle_delattr,
            consts.HANDLE_SETATTR: cls._handle_setattr,
            consts.HANDLE_CALL: cls._handle_call,
            consts.HANDLE_CALLATTR: cls._handle_callattr,
            consts.HANDLE_REPR: cls._handle_repr,
            consts.HANDLE_STR: cls._handle_str,
            consts.HANDLE_CMP: cls._handle_cmp,
            consts.HANDLE_HASH: cls._handle_hash,
            consts.HANDLE_INSTANCECHECK: cls._handle_instancecheck,
            consts.HANDLE_DIR: cls._handle_dir,
            consts.HANDLE_PICKLE: cls._handle_pickle,
            consts.HANDLE_DEL: cls._handle_del,
            consts.HANDLE_INSPECT: cls._handle_inspect,
            consts.HANDLE_BUFFITER: cls._handle_buffiter,
            consts.HANDLE_OLDSLICING: cls._handle_oldslicing,
            consts.HANDLE_CTXEXIT: cls._handle_ctxexit,
            consts.HANDLE_ABOUT_TO_CLOSE: cls._handle_about_to_close,
        }

    def _handle_about_to_close(self):  # request handler
        self._pre_cleanup()


class InstrumentService(rpyc.Service):

    _protocol = InstrumentConnection

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def on_about_to_disconnect(self, conn):
        """called when the connection had already terminated for cleanup
        (must not perform any IO on the connection)"""
        pass


class VoidInstrumentService(InstrumentService):
    """void service - an do-nothing service"""
    __slots__ = ()


rpyc.Connection = InstrumentConnection
rpyc.Service = InstrumentService
rpyc.VoidService = VoidInstrumentService


class InstrumentServer(InstrumentService):
    """RPyC provider that loads lantz devices and exposes them to the remote
    client
    """

    def __init__(self, config_file=None, mongo_addr=None):
        super().__init__()
        # if the config file isn't specified, get it from the meta-config
        if not config_file:
            config_file = load_meta_config(SERVER_META_CONFIG_PATH)
        # lantz devices
        self._devs = {}
        # configuration
        self.config = {}
        self.config_file = None
        # server settings
        self.port = None
        # rpyc server object
        self._rpyc_server = None
        # mongodb
        self.db_name = None
        self.mongo_addr = None
        self.mongo_client = None
        self.db = None

        # storage container for hook methods that update mongodb feats when
        # lantz feats are changed - this is required because pysignal uses
        # weakrefs, so any local functions you pass to it will fail when they
        # go out of scope
        # TODO replace with partials
        self.feat_hook_functions = {}
        self.dictfeat_hook_functions = {}

        # start everything
        self.reload_config(config_file)
        self.reload_server_config()
        self.connect_mongo(mongo_addr)
        self.reload_devices()
        self.start_server()

    def __getattr__(self, name):
        """Allow the user to access the driver objects directly using
        e.g. inserv.fake_tcpip_sg.amplitude notation
        """
        if name in self._devs:
            return self._devs[name]
        else:
            return self.__getattribute__(name)
            #raise AttributeError('\'{}\' object has no attribute \'{}\''.\
            #            format(self.__class__.__name__, name))


    def on_connect(self, conn):
        """Called when a client connects to the RPyC server"""
        logger.info('client [{}] connected'.format(conn))

    def on_about_to_disconnect(self, conn):
        logger.info('client [{}] about to disconnect'.format(conn))
        for device_name, device in self._devs.items():
            # iterate over all feats and dictfeats
            for attr_name, attr in list(device._lantz_feats.items()) + list(device._lantz_dictfeats.items()):
                if isinstance(attr_name, DictPropertyNameKey):
                    # filter out weird dictfeat feats
                    continue
                # access the PySignal slots
                attr_slots = getattr(device, attr_name + '_changed')._slots
                for slot in attr_slots:
                    if isinstance(slot, functools.partial) and slot.__name__ == 'InstrumentManager_getattr_func':
                        getattr(device, attr_name + '_changed').disconnect(slot)


    def on_disconnect(self, conn):
        """Called when a client disconnects from the RPyC server"""
        logger.info('client [{}] disconnected'.format(conn))

        # when an RPyC client disconnects, there are dangling netrefs
        # left over in PySignal _slots that point to objects on the client
        # side which are now inaccessible
        # this block detects and removes the references by attempting to
        # access them with a try/except
        # TODO this logic should really be implemented in PySignal during emit()
        for device_name, device in self._devs.items():
            # iterate over all feats and dictfeats
            for attr_name, attr in list(device._lantz_feats.items()) + list(device._lantz_dictfeats.items()):
                if isinstance(attr_name, DictPropertyNameKey):
                    # filter out weird dictfeat feats
                    continue
                # access the PySignal slots
                attr_slots = getattr(device, attr_name + '_changed')._slots
                for index, slot in enumerate(attr_slots):
                    try:
                        # trying to compare slot to something will force RPyC
                        # to attempt to retrieve it
                        if slot == None:
                            pass
                    except EOFError:
                        del attr_slots[index]


    def restart(self, config_file=None, mongo_addr=None):
        """Restart the server AND reload the config file and all devices"""
        logger.info('restarting...')
        self.reload_config(config_file)
        self.reload_server_config()
        self.disconnect_mongo()
        self.connect_mongo(mongo_addr)
        self.reload_devices()
        self.reload_server()


    def reload_server_config(self):
        """Reload RPyC server settings from the config"""
        self.port, _ = get_config_param(self.config, ['port'])


    def connect_mongo(self, mongo_addr=None):
        """Config and connect to the mongodb database"""
        self.db_name = MONGO_SERVERS_KEY.format(self.port)
        if mongo_addr:
            self.mongo_addr = mongo_addr
        else:
            self.mongo_addr,_ = get_config_param(self.config, [CONFIG_MONGO_ADDR_KEY])
        logger.info('connecting to mongodb server [{}]...'.format(self.mongo_addr))
        self.mongo_client = pymongo.MongoClient(mongo_addr, replicaset=MONGO_RS, serverSelectionTimeoutMS=MONGO_CONNECT_TIMEOUT)
        self.db = self.mongo_client[self.db_name]
        try:
            self.mongo_client.drop_database(self.db_name)
        except Exception as exc:
            raise InstrumentServerError('Failed connecting to mongodb [{}]'.format(self.mongo_addr), exception=exc) from None
        logger.info('connected to mongodb server [{}]'.format(self.mongo_addr))


    def disconnect_mongo(self):
        """Disconnect from the mongodb database"""
        if self.mongo_client:
            # remove the database entry from mongo
            self.mongo_client.drop_database(self.db_name)
            # disconnect
            self.mongo_client.close()


    def add_device(self, dev_name):
        """Add and initialize a device"""

        # load the device parameters from the config file

        # first try getting the driver's lantz class
        try:
            dev_lantz_class, dev_class_cfg_file = get_config_param(self.config,
                                        [CONFIG_SERVER_DEVICES, dev_name, \
                                        CONFIG_SERVER_DEVICE_LANTZ_CLASS])
            try:
                dev_class = load_class_from_str('lantz.drivers.' + \
                                                dev_lantz_class)
            except Exception as exc:
                raise InstrumentServerError('The specified lantz driver '
                    '[{}] for device [{}] couldn\'t be loaded'.\
                    format(dev_lantz_class, dev_name), exception=exc) from None
        except EntryNotFoundError:
            # if the lantz class isn't defined, try getting an absolute file
            # path and class name
            try:
                dev_class_file_str, dev_class_file_cfg_file = \
                                    get_config_param(self.config,
                                            [CONFIG_SERVER_DEVICES, dev_name,
                                            CONFIG_SERVER_DEVICE_CLASS_FILE])
                dev_class_name, _ = get_config_param(self.config,
                                            [CONFIG_SERVER_DEVICES, dev_name,
                                            CONFIG_SERVER_DEVICE_CLASS_NAME])
            except EntryNotFoundError as exc:
                raise InstrumentServerError('The device [{}] didn\'t '
                    'contain an entry for either a lantz class "{}" or a file '
                    'path "{}" / class name "{}" to define it\'s driver type'.\
                    format(dev_name, CONFIG_SERVER_DEVICE_LANTZ_CLASS,
                            CONFIG_SERVER_DEVICE_CLASS_FILE,
                            CONFIG_SERVER_DEVICE_CLASS_NAME), exception=exc) from None
            dev_class_path = Path(dev_class_file_str)

            # resolve relative paths
            if not dev_class_path.is_absolute():
                dev_class_path = Path(dev_class_file_cfg_file).parent / dev_class_path
            dev_class_path = dev_class_path.resolve()

            try:
                dev_class = load_class_from_file(dev_class_path, dev_class_name)
            except Exception as exc:
                raise InstrumentServerError('The specified class [{}] '
                    'from file [{}] for device [{}] couldn\'t be loaded'.\
                    format(dev_class_name, dev_class_path, dev_name), exception=exc) from None

        dev_args, _ = get_config_param(self.config,
                                    [CONFIG_SERVER_DEVICES, dev_name, 'args'])
        dev_kwargs, _ = get_config_param(self.config,
                                    [CONFIG_SERVER_DEVICES, dev_name, 'kwargs'])

        # get an instance of the device
        try:
            self._devs[dev_name] = dev_class(*dev_args, **dev_kwargs)
        except Exception as exc:
            raise InstrumentServerError('Failed to get instance of device {} of class {}'.\
                                        format(dev_name, dev_class), exception=exc) from None

        # collect all of the lantz feature attributes
        feat_attr_list = []
        for feat_name, feat in list(dev_class._lantz_feats.items()) + list(dev_class._lantz_dictfeats.items()):
            if isinstance(feat_name, DictPropertyNameKey):
                # filter out weird dictfeat feats
                continue

            attrs = copy.deepcopy(feat.__dict__['_config'])
            if isinstance(feat, DictFeat) and 'keys' in feat.__dict__:
                attrs['keys'] = feat.__dict__['keys']

            if attrs['limits']:
                if len(attrs['limits']) == 1:
                    limits = [0, attrs['limits'][0]]
                else:
                    limits = attrs['limits']
            else:
                limits = None

            if attrs['values']:
                values = list(attrs['values'])
            else:
                values = None

            if 'keys' in attrs and attrs['keys']:
                keys = list(attrs['keys'])
            else:
                keys = None

            if 'keys' in attrs and attrs['keys'] and isinstance(feat, DictFeat):
                value = [None] * len(attrs['keys'])
            else:
                value = None

            feat_attr_list.append({
                'name':     feat_name,
                'type':     'dictfeat' if isinstance(feat, DictFeat) \
                                            else 'feat',
                'readonly': feat.__dict__['fset'] is None,
                'units':    attrs['units'],
                'limits':   limits,
                'values':   values,
                'keys':     keys,
                'value':    value
            })

            # add a custom hook for updating mongodb whenever the feat/dictfeat is written
            if isinstance(feat, DictFeat):
                def update_mongo_dictfeat(value, old_value, key, attr=feat_name, keys=keys):
                    logger.debug('{}[{}]: {} -> {}'.format(attr, key, old_value, value))
                    if isinstance(value, Quantity):
                        value = value.to(self._devs[dev_name]._lantz_dictfeats[attr]._kwargs['units']).m
                    self.db[dev_name].update_one({'name': attr},
                                                 {'$set': {'value.{}'.format(keys.index(key)): value}},
                                                 upsert=True)

                self.dictfeat_hook_functions[feat_name] = update_mongo_dictfeat
                getattr(self._devs[dev_name], feat_name + '_changed').connect(self.dictfeat_hook_functions[feat_name])
            else:
                def update_mongo_feat(value, old_value, attr=feat_name):
                    logger.debug('{}: {} -> {}'.format(attr, old_value, value))
                    if isinstance(value, Quantity):
                        value = value.to(self._devs[dev_name]._lantz_feats[attr]._kwargs['units']).m
                    self.db[dev_name].update_one({'name': attr},
                                                 {'$set': {'value': value}},
                                                 upsert=True)

                self.feat_hook_functions[feat_name] = update_mongo_feat
                getattr(self._devs[dev_name], feat_name + '_changed').connect(self.feat_hook_functions[feat_name])

        for action_name, action in dev_class._lantz_actions.items():
            feat_attr_list.append({'name' : action_name, 'type' : 'action'})

        self.db[dev_name].drop()
        # add all of the lantz feature attributes to the database
        self.db[dev_name].insert_many(feat_attr_list)

        # initialize the device
        try:
            self._devs[dev_name].initialize()
        except Exception as exc:
            logger.error(exc)
            self._devs.pop(dev_name)
            logger.error('device [{}] initialization sequence failed'.\
                            format(dev_name))
            return

        logger.info('added device [{}] with args: {} kwargs: {}'.\
                        format(dev_name, dev_args, dev_kwargs))


    def del_device(self, dev_name):
        """Remove and finalize a device"""
        try:
            self._devs.pop(dev_name).finalize()
        except Exception as exc:
            raise InstrumentServerError('Failed deleting device [{}]'.\
                                        format(dev_name), exception=exc) from None
        logger.info('deleted [{}]'.format(dev_name))


    def reload_device(self, dev_name):
        """Remove a device, then reload it from the stored config"""
        if dev_name in self._devs:
            self.del_device(dev_name)
        self.add_device(dev_name)


    def reload_devices(self):
        """Reload all devices"""
        devs, _ = get_config_param(self.config, ['devices'])
        for dev_name in devs:
            self.reload_device(dev_name)
        logger.info('reloaded all devices')


    def reload_config(self, config_file=None):
        """Reload the config files"""
        # update the config with a new file if one was passed as argument
        if config_file:
            self.config_file = config_file
        # reload the config dictionary
        self.config = load_config(self.config_file)
        logger.info('loaded config files {}'.\
                        format(list(self.config.keys())))


    def reload_server(self):
        """Restart the RPyC server"""
        self.stop_server()
        self.start_server()


    def start_server(self):
        """Start the RPyC server"""
        if self._rpyc_server:
            logger.warning('can\'t start the rpyc server because one '
                            'is already running')
            return
        thread = threading.Thread(target=self._rpyc_server_thread)
        thread.start()
        # wait for the server to start
        while not (self._rpyc_server and self._rpyc_server.active):
            time.sleep(0.1)


    def stop_server(self):
        """Stop the RPyC server"""
        if not self._rpyc_server:
            logger.warning('can\'t stop the rpyc server because there '
                            'isn\'t one running')
            return
        logger.info('stopping RPyC server...')
        self._rpyc_server.close()
        # wait for the server to stop
        while self._rpyc_server.active:
            time.sleep(0.1)
        RPYC_SERVER_STOP_EVENT.wait()
        RPYC_SERVER_STOP_EVENT.clear()
        self._rpyc_server = None


    def _rpyc_server_thread(self):
        """Thread for running the RPyC server asynchronously"""
        logger.info('starting RPyC server...')
        self._rpyc_server = ThreadedServer(self,
                        hostname='127.0.0.1', port=self.port,
                        protocol_config={'allow_pickle' : True,
                                    'allow_all_attrs' : True,
                                    'allow_setattr' : True,
                                    'allow_delattr' : True,
                                    'sync_request_timeout' : RPYC_SYNC_TIMEOUT})
        self._rpyc_server.start()
        logger.info('RPyC server stopped')
        RPYC_SERVER_STOP_EVENT.set()



import logging
import sys

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QApplication

from nspyre.misc import nspyre_init_logger

from .app import NSpyreApp
from .main_window import NSpyreMainWindow

from .colors import colors
from .data_handling import save_data
from .widgets.plotting import HeatmapPlotWidget, LinePlotWidget
from .widgets.views import Plot1D, Plot2D, PlotFormatInit, PlotFormatUpdate

__all__ = [
    'HeatmapPlotWidget',
    'LinePlotWidget',
    'Plot1D',
    'Plot2D',
    'PlotFormatInit',
    'PlotFormatUpdate',
    'colors',
    'save_data'
]

logger = logging.getLogger(__name__)


def main(args=None):
    """Entry point for the application script."""

    nspyre_init_logger(logging.INFO)
    if hasattr(Qt, 'AA_EnableHighDpiScaling'):
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
        QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
    app = NSpyreApp(['NSpyre', sys.argv])
    main_window = NSpyreMainWindow()
    sys.exit(app.exec())



"""The class defining the NSpyre Application.

The NSpyreApp class is the entry point for the NSpyre software application.
It is defined by the QtWidgets.QApplication subclass from Qt and defines
all the parameters for setting up the Qt GUI.

From the Qt documentation:
The QApplication class manages the GUI application's control flow and main settings.
It contains the main event loop, where all events from the window system and other
sources are processed and dispatched. It also handles the application's initialization
and finalization, and provides session management. It also handles most system-wide
and application-wide settings. For any GUI application that uses Qt, there is
precisely one QApplication object, no matter whether the application has 0, 1, 2 or
more windows at any time.

  Typical usage example:

  app = NSpyreApp([sys.argv])
  window = main_window.NSpyreMainWindow()
  sys.exit(app.exec())

Copyright (c) 2020, Alexandre Bourassa, Michael Solomon, Jacob Feder
All rights reserved.

This work is licensed under the terms of the 3-Clause BSD license.
For a copy, see <https://opensource.org/licenses/BSD-3-Clause>.
"""
from pathlib import Path

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QColor, QFont, QIcon, QPalette
from PyQt5.QtWidgets import QApplication, QStyleFactory

HERE = Path(__file__).parent


class NSpyreApp(QApplication):
    """The Qt QtWidgets.QApplication object for launching NSpyre.

    This is the class you need to instantiate for starting a Qt GUI.
    This has the additional settings and parameters determining the
    appearance of the GUI application.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setApplicationName('NSpyre')
        icon_path = HERE / 'images/favicon.ico'
        self.setWindowIcon(QIcon(str(icon_path)))
        self.set_theme()

    def set_theme(self):
        """A method containing the appearance settings for NSpyre."""
        fusion = QStyleFactory.create('Fusion')
        self.setStyle(fusion)
        dark = QColor(53, 53, 53) #(61, 64, 62) (196,201,201)
        palette = QPalette()
        palette.setColor(QPalette.Window, dark)
        palette.setColor(QPalette.WindowText, Qt.white)
        palette.setColor(QPalette.Base, QColor(25, 25, 25))
        palette.setColor(QPalette.AlternateBase, dark)
        palette.setColor(QPalette.ToolTipText, Qt.white)
        palette.setColor(QPalette.Text, Qt.white)
        palette.setColor(QPalette.Button, dark)
        palette.setColor(QPalette.ButtonText, Qt.white)
        palette.setColor(QPalette.BrightText, Qt.red)
        palette.setColor(QPalette.Link, QColor(42, 130, 218))
        palette.setColor(QPalette.Highlight, QColor(42, 130, 218))
        palette.setColor(QPalette.HighlightedText, Qt.black)
        self.setPalette(palette)
        style_sheet = (HERE / 'style.qss').read_text()
        self.setStyleSheet(style_sheet)
        self.setFont(QFont('Helvetica [Cronyx]'))



from nspyre.spyrelet.spyrelet import (
    Spyrelet,
    SpyreletLauncher,
    LOADED_SPYRELETS,
    load_spyrelet_class,
    load_spyrelet,
    load_all_spyrelets,
    unload_spyrelet,
    unload_all_spyrelets,
    reload_spyrelet,
    reload_all_spyrelets
)

__all__ = [
    'Spyrelet',
    'LOADED_SPYRELETS',
    'load_spyrelet',
    'load_all_spyrelets',
    'unload_spyrelet',
    'unload_all_spyrelets',
    'reload_spyrelet',
    'reload_all_spyrelets'
]



from PyQt5 import QtWidgets

from ..data_handling import save_data


class Save_Widget(QtWidgets.QWidget):

    def __init__(self, spyrelet, parent=None):
        super().__init__(parent=parent)
        self.spyrelet = spyrelet
        self.filename = None
        self.init_ui()
        return

    def init_ui(self):
        # Create file selection widget
        file_w = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout()
        file_w.setLayout(layout)
        self.filename_label = QtWidgets.QLabel('No filename currently selected')
        self.select_repository = QtWidgets.QPushButton('Select repository...')
        self.select_repository.clicked.connect(self.select_repository_dialog)
        layout.addWidget(self.select_repository)
        layout.addWidget(self.filename_label)

        # Entry widget
        entry_w = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout()
        entry_w.setLayout(layout)
        label = QtWidgets.QLabel('Entry name: ')
        self.dataset_name = QtWidgets.QLineEdit()
        layout.addWidget(label)
        layout.addWidget(self.dataset_name)
        self.dataset_name.setText(self.spyrelet.__class__.__name__)

        # Other widget
        self.dataset_description = QtWidgets.QTextEdit()
        self.save_btn = QtWidgets.QPushButton('Save')
        self.save_btn.clicked.connect(self.save)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(QtWidgets.QLabel('Save widget for {}'.format(self.spyrelet.name)))
        layout.addWidget(file_w)
        layout.addWidget(entry_w)
        layout.addWidget(QtWidgets.QLabel('Description:'))
        layout.addWidget(self.dataset_description)
        layout.addWidget(self.save_btn)
        self.setLayout(layout)
        return

    def select_repository_dialog(self):
        filename, other = QtWidgets.QFileDialog.getSaveFileName(None, 'Save spyrelet to...', '', 'JSON files (*.json)')
        self.filename = filename
        self.filename_label.setText(filename)
        return

    def save(self):
        description = self.dataset_description.toPlainText()
        name = self.dataset_name.text()
        if self.filename:
            save_data(self.spyrelet, self.filename, name=name, description=description)
            print('Data for {} was saved under {}'.format(self.spyrelet.name, self.filename))



import os

from enum import Enum

from PyQt5 import QtWidgets, QtCore

__package__ = 'nspyre.gui.widgets'

from ..app import NSpyreApp


class SplitterOrientation(Enum):

    vertical_left_button = 0
    vertical_right_button = 1
    horizontal_top_button = 2
    horizontal_bottom_button = 3


class Splitter(QtWidgets.QSplitter):

    def __init__(self, main_w, side_w, orientation=None, parent=None):
        if orientation is None:
            orientation = SplitterOrientation.vertical_left_button

        if orientation in [SplitterOrientation.vertical_left_button,
                           SplitterOrientation.vertical_right_button]:
            super().__init__(QtCore.Qt.Horizontal, parent=parent)
        else:
            super().__init__(QtCore.Qt.Vertical, parent=parent)
        horizontal = int(orientation.value >= 2)
        if not orientation.value % 2:
            # main widget on container 1
            c1_hover = SplitterHoverArea(self, orientation=orientation)
            c2_hover = None
            initial_size = [1, 0]
        else:
            c1_hover = None
            c2_hover = SplitterHoverArea(self, orientation=orientation)
            initial_size = [0, 1]
        c1 = Container(main_w, c1_hover)
        c2 = Container(side_w, c2_hover)
        self.addWidget(c1)
        self.addWidget(c2)
        self.orientation = orientation
        self.setSizes(initial_size)
        self.showMaximized()
        return

    def setSizes(self, sizes):
        super().setSizes(sizes)
        button = self.widget((self.orientation.value % 2)).collapse_area.button
        if ((not self.orientation.value % 2 and not sizes[1]) or
            (self.orientation.value % 2 and not sizes[0])):
            button.closed = True
            button.on_splitter_moved(None, None)
        else:
            button.closed = False
            button.on_splitter_moved(None, None)
        return


class SplitterHoverArea(QtWidgets.QWidget):

    width = 30
    height = 60

    def __init__(self, splitter, orientation=None, parent=None):
        if orientation is None:
            orientation = SplitterOrientation.vertical_left_button
        super().__init__(parent=parent)
        if orientation.value % 2:
            self.even = False
        else:
            self.even = True
        if orientation.value >= 2:
            # horizontal orientation
            self.width, self.height = self.height, self.width
            self.start_x = 0
            self.start_y = 0
            if self.even:
                self.end_x = self.height
            else: self.end_x = -self.height
            self.end_y = 0
        else:
            self.start_x = 0
            self.start_y = 0
            self.end_x = 0
            if self.even:
                self.end_y = self.width
            else:
                self.end_y = -self.width
        self.setFixedSize(self.width, self.height)
        self.orientation = orientation
        self.button = SplitterHoverButton(splitter, parent=self)
        self.on_hover(False, duration=0)
        return

    def enterEvent(self, ev):
        self.on_hover(True)
        return

    def leaveEvent(self, ev):
        self.on_hover(False)
        return

    def on_hover(self, entering, duration=200):
        animation = QtCore.QPropertyAnimation(self)
        animation.setDuration(duration)
        animation.setTargetObject(self.button)
        animation.setPropertyName(b'pos')
        animation.setEasingCurve(QtCore.QEasingCurve.Linear)

        if entering:
            start = QtCore.QPoint(self.end_y, self.end_x)
            end = QtCore.QPoint(self.start_y, self.start_x)
        else:
            start = QtCore.QPoint(self.start_y, self.start_x)
            end = QtCore.QPoint(self.end_y, self.end_x)

        animation.setStartValue(start)
        animation.setEndValue(end)
        animation.start()
        return


class SplitterHoverButton(QtWidgets.QPushButton):

    button_style = '''
        background-image: url({image});
        background-repeat: no-repeat;
        background-position: center;
        background-color: rgb(53, 53, 53);
        width: {width}px;
        height: {height}px;
        border-{corner1}-radius: 30px;
        border-{corner2}-radius: 30px;
        border-color: rgb(20, 20, 20);
        border-style: outset;
        border-width: {border_width};
    '''

    def __init__(self, splitter, parent=None):
        super().__init__(parent=parent)
        hover_area = self.parent()
        self.closed = False
        if hover_area.orientation.value % 2:
            self.even = False
        else:
            self.even = True

        image_path = os.path.join(os.path.dirname(__file__),'..\\images\\')
        image_path = image_path.replace("\\","/")

        if hover_area.orientation.value >= 2:
            # horizontal orientation
            self.closed_image = image_path + '/{}_arrow.png'.format('up' if self.even else 'down')
            self.opened_image = image_path + '/{}_arrow.png'.format('down' if self.even else 'up')
            self.width = 60
            self.height = 30
            self.button_params = {
                'image': self.opened_image,
                'width': 60,
                'height': 30,
                'corner1': 'top-left',
                'corner2': 'top-right',
                'border_width': '0.5px 0.5px 0 0.5px',
            }
            if not self.even:
                self.button_params['corner1'] = 'bottom-left'
                self.button_params['corner2'] = 'bottom-right'
                self.button_params['border_width'] = '0 0.5px 0.5px 0.5px'

        else:
            self.closed_image = image_path + '/{}_arrow.png'.format('left' if self.even else 'right')
            self.opened_image = image_path + '/{}_arrow.png'.format('right' if self.even else 'left')
            self.width = 30
            self.height = 60
            self.button_params = {
                'image': self.opened_image,
                'width': 30,
                'height': 60,
                'corner1': 'top-left',
                'corner2': 'bottom-left',
                'border_width': '0.5px 0 0.5px 0.5px',
            }
            if not self.even:
                self.button_params['corner1'] = 'top-right'
                self.button_params['corner2'] = 'bottom-right'
                self.button_params['border_width'] = '0.5px 0.5px 0.5px 0'


        self.style = self.button_style.format(**self.button_params)
        self.setStyleSheet(self.style)
        self.setFixedSize(self.width, self.height)
        self.setCursor(QtCore.Qt.PointingHandCursor)
        self.splitter = splitter
        self.maximize = 400
        self.minimize = 1
        self.splitter.splitterMoved.connect(self.on_splitter_moved)
        self.clicked.connect(self.on_clicked)
        return

    def on_clicked(self):
        self.setStyleSheet(self.style)
        if self.closed:
            size = [self.maximize, self.minimize]
        else:
            size = [1,0]
        if not self.even: size[0], size[1] = size[1], size[0]
        self.splitter.setSizes(size)
        self.on_splitter_moved(None, None)
        return

    def on_splitter_moved(self, pos, idx):
        sizes = self.splitter.sizes()
        if not self.even: sizes[0], sizes[1] = sizes[1], sizes[0]
        hover_area = self.parent()
        if not sizes[1]:
            self.button_params['image'] = self.closed_image
            self.closed = True
        else:
            self.button_params['image'] = self.opened_image
            self.closed = False
            self.maximize = sizes[0]
            self.minimize = sizes[1]
        self.style = self.button_style.format(**self.button_params)
        self.setStyleSheet(self.style)
        return

    def resizeEvent(self, ev):
        ev.ignore()
        return


class Container(QtWidgets.QWidget):

    def __init__(self, w, collapse=None, parent=None):
        super().__init__(parent=parent)
        self.layout = QtWidgets.QHBoxLayout(self)
        self.layout.addWidget(w)
        self.setLayout(self.layout)
        self.layout.setContentsMargins(0,0,0,0)
        if collapse is not None:
            self.collapse_area = collapse
            self.collapse_area.setParent(self)
        else:
            self.collapse_area = None
        return

    def resizeEvent(self, ev):
        if self.collapse_area is not None:
            if self.collapse_area.orientation.value % 2:
                even = False
            else:
                even = True
            if self.collapse_area.orientation.value >= 2:
                # horizontal orientation
                size = ev.size()
                b_size = self.collapse_area.size()
                w, h = size.width(), size.height()
                b_w, b_h = b_size.width(), b_size.height()
                if even:
                    self.collapse_area.move((w - b_w) / 2, h - b_h)
                else:
                    self.collapse_area.move((w - b_w) / 2, 0)
            else:
                size = ev.size()
                b_size = self.collapse_area.size()
                w, h = size.width(), size.height()
                b_w, b_h = b_size.width(), b_size.height()
                if even:
                    self.collapse_area.move(w - b_w, (h - b_h) / 2)
                else:
                    self.collapse_area.move(0, (h - b_h) / 2)


def main():
    app = NSpyreApp([])

    layout = QtWidgets.QGridLayout()

    for orientation in range(4):
        s_orientation = SplitterOrientation(orientation)
        left_item = QtWidgets.QLabel('Left')
        right_item = QtWidgets.QLabel('Right')
        s = Splitter(left_item, right_item, orientation=s_orientation)
        layout.addWidget(s, orientation // 2, orientation % 2)

    container = QtWidgets.QWidget()

    container.setLayout(layout)
    container.show()
    app.exec_()
    return


if __name__ == '__main__':
    main()



from pint.util import infer_base_unit
from pyqtgraph import SpinBox as _SpinBox

from nspyre.definitions import Q_


class SpinBox(_SpinBox):

    def __init__(self, parent=None, value=1.0, unit=None, **kwargs):
        self.unit = unit
        self.base_unit = None
        kwargs['dec'] = kwargs.pop('dec', True)
        kwargs['minStep'] = kwargs.pop('minStep', 0.1)
        kwargs['decimals'] = kwargs.pop('decimals', 8)
        # kwargs['compactHeight'] = False
        if self.unit is not None:
            q = Q_(unit)
            base_units = infer_base_unit(q)
            base_unit = '*'.join('{} ** {}'.format(u, p) for u, p in base_units.items())
            self.base_unit = base_unit if not base_unit == '' else 'dimensionless'
            q_base = Q_(self.base_unit)
            factor = (q / q_base).to_base_units().m
            opts = {
                'suffix': '{0.units:~}'.format(q_base),
                'siPrefix': True,
            }
            kwargs.update(opts)
            value *= factor
        super().__init__(parent=parent, value=value, **kwargs)
        self.setMaximumHeight(1e6)
        return

    def getValue(self):
        if self.base_unit is None:
            return super().value()
        else:
            return Q_(super().value(), self.base_unit)

    def unit_value(self):
        val = super().value()
        uval = Q_(val, self.base_unit) if self.base_unit is not None else val
        return uval

    def setValue(self, value=None, **kwargs):
        if isinstance(value, Q_):
            value = value.to_base_units().m
        super().setValue(value=value, **kwargs)



from setuptools import setup, find_packages
import codecs
import pathlib
import re


here = pathlib.Path(__file__).parent.resolve()


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with codecs.open(pathlib.PurePath(here, *parts), "rb", "utf-8") as f:
        return f.read()


def find_version(*file_paths):
    """
    Build a path from *file_paths* and search for a ``__version__``
    string inside.
    """
    version_file = read(*file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


meta_path = pathlib.PurePath('src', 'nspyre', '__init__.py')
version = find_version(meta_path)

long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name='nspyre',
    version=version,
    license='BSD 3-Clause License',
    description='Networked Scientific Python Research Environment',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/nspyre-org/nspyre',
    author='Alexandre Bourassa',
    author_email='abourassa@uchicago.edu',
    maintainer='Michael Solomon',
    maintainer_email='msolo@uchicago.edu',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: IPython',
        'Framework :: Jupyter',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: User Interfaces',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Logging',
    ],
    keywords='nspyre, measurement toolkit, experimentation platform, physics, science, research',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    zip_safe=False,
    python_requires='>=3.8, <4',
    install_requires=[
        # SciPy
        'numpy>=1.19.1',
        'scipy>=1.5.2',
        'pandas>=1.1.2',
        # MongoDB
        'mongodb<6.0.0',
        'pymongo>=3.10.1',
        # Qt
        'pyqt5>=5.12.3',
        'pyqtgraph>=0.11.0',
        'qscintilla>=2.11.2',
        # VISA
        'pyvisa>=1.10.1',
        # Lantz
        'pint>=0.15',
        'pimpmyclass>=0.4.3',
        'lantzdev>=0.5.2',
        # Utilities
        'parse>=1.18.0',
        'tqdm>=4.49.0',
        'rpyc>=4.1.5',
    ],
    extras_require={
        'dev': [
            'pytest>=6.1.2',
            'pytest-cov',
            'psutil>=5.7.3',
        ]
    },
    test_requires=[
        'pytest>=6.1.2',
        'pytest-cov',
        'psutil>=5.7.3',
    ],
    test_suite='tests',
    entry_points={
        'console_scripts': [
            'nspyre=nspyre.gui:main',
            'nspyre-config=nspyre.config.config_cli:main',
            'nspyre-mongodb=nspyre.mongodb:main',
            'nspyre-inserv=nspyre.inserv:main',
        ],
    },
    project_urls={
        'Bug Reports': 'https://github.com/nspyre-org/nspyre/issues',
        'Source': 'https://github.com/nspyre-org/nspyre/',
    },
    include_package_data=True,
    options={'bdist_wheel': {'universal': '1'}},
)



from PyQt5.Qsci import QsciLexerPython, QsciScintilla
from PyQt5 import QtCore, QtGui


class Scintilla_Code_Editor(QsciScintilla):
    request_run_signal = QtCore.pyqtSignal()

    def keyPressEvent(self, ev):
        if ev.modifiers() & QtCore.Qt.ControlModifier and ev.key() == QtCore.Qt.Key_Slash:
            # Commenting/Uncommenting on Ctr+/
            if self.hasSelectedText():
                isSelection = True
                selection = list(self.getSelection())
                start, _, end, _ = selection
                end += 1
            else:
                isSelection = False
                cursor_pos = list(self.getCursorPosition())
                start, _ = cursor_pos
                end = start + 1
            self.setSelection(start,0,end,0)
            text = self.selectedText()
            split = text.split('\n')
            if all([line=='' or line[0]=='#' for line in split]): #Need to uncomment if line starts with '#'
                new_split = ['' if line=='' else line[1:] for line in split]
            else: #Need to comment
                new_split = ['' if line=='' else '#'+line for line in split]
            new_text = '\n'.join(new_split)
            self.removeSelectedText()
            self.insert(new_text)

            if isSelection:
                selection[3] += len(new_text)-len(text)
                self.setSelection(*selection)
            else:
                cursor_pos[1] += len(new_text)-len(text)
                self.setCursorPosition(*cursor_pos)
            ev.accept()
        elif ev.modifiers() & QtCore.Qt.ShiftModifier and ev.key() == QtCore.Qt.Key_Return:
            self.request_run_signal.emit()
            
        else:
            super().keyPressEvent(ev)


class Monokai_Python_Lexer(QsciLexerPython):
    def __init__(self, editor, *args):
        super().__init__(editor, *args)

        #Margins
        editor.setMarginLineNumbers(1, True)
        editor.setMarginWidth(1, '99999')
        editor.setMarginsBackgroundColor(QtGui.QColor("#272822"))
        editor.setMarginsForegroundColor(QtGui.QColor('#75715e'))

        #Cursor
        editor.ensureCursorVisible()
        editor.setCaretForegroundColor(QtGui.QColor('white'))
        editor.setCaretWidth(2)

        #Indentation
        editor.setAutoIndent(True)
        editor.setIndentationsUseTabs(False)
        editor.setTabWidth(4)


        self.setDefaultPaper(QtGui.QColor("#272822"))
        font = QtGui.QFont('Courier New',12)
        self.setFont(font)

        self.setIndentationWarning(QsciLexerPython.Inconsistent)
        self.setHighlightSubidentifiers(False)

        self.setColor(QtGui.QColor('#9081ff'), QsciLexerPython.Number) ## Purple
        self.setColor(QtGui.QColor('#66d9ef'), QsciLexerPython.HighlightedIdentifier )## Blue
        self.setColor(QtGui.QColor('#e2266d'), QsciLexerPython.Keyword) ## keyword red
        self.setColor(QtGui.QColor('white'), QsciLexerPython.Operator) ## White
        self.setColor(QtGui.QColor('#75715e'), QsciLexerPython.Comment) ## Comment Gray

        self.setColor(QtGui.QColor('#e6db5a'), QsciLexerPython.DoubleQuotedString) ## String Yellow
        self.setColor(QtGui.QColor('#e6db5a'), QsciLexerPython.SingleQuotedString) ## String Yellow
        self.setColor(QtGui.QColor('#e6db5a'), QsciLexerPython.TripleSingleQuotedString) ## String Yellow
        self.setColor(QtGui.QColor('#e6db5a'), QsciLexerPython.TripleDoubleQuotedString) ## String Yellow
        
        self.setColor(QtGui.QColor('#a6e22e'), QsciLexerPython.ClassName) ## Green
        self.setColor(QtGui.QColor('#a6e22e'), QsciLexerPython.FunctionMethodName) ## Green
        self.setColor(QtGui.QColor('#a6e22e'), QsciLexerPython.Decorator) ## Green

    def keywords(self, s):
        # if s == 1:
        #     return super().keywords(s) + '+ - * / =='
        if s == 2:
            special_func = '__abs__ __add__ __and__ __call__ __class__ __cmp__ __coerce__ __complex__ __contains__ __del__ __delattr__ __delete__ __delitem__ __delslice__ __dict__ __div__ __divmod__ __eq__ __float__ __floordiv__ __ge__ __get__ __getattr__ __getattribute__ __getitem__ __getslice__ __gt__ __hash__ __hex__ __iadd__ __iand__ __idiv__ __ifloordiv__ __ilshift__ __imod__ __imul__ __index__ __init__ __instancecheck__ __int__ __invert__ __ior__ __ipow__ __irshift__ __isub__ __iter__ __itruediv__ __ixor__ __le__ __len__ __long__ __lshift__ __lt__ __metaclass__ __mod__ __mro__ __mul__ __ne__ __neg__ __new__ __nonzero__ __oct__ __or__ __pos__ __pow__ __radd__ __rand__ __rcmp__ __rdiv__ __rdivmod__ __repr__ __reversed__ __rfloordiv__ __rlshift__ __rmod__ __rmul__ __ror__ __rpow__ __rrshift__ __rshift__ __rsub__ __rtruediv__ __rxor__ __set__ __setattr__ __setitem__ __setslice__ __slots__ __str__ __sub__ __subclasscheck__ __truediv__ __unicode__ __weakref__ __xor__'
            built_in = 'abs divmod input open staticmethod all enumerate int ord str any eval isinstance pow sum basestring execfile issubclass print super bin file iter property tuple bool filter len range type bytearray float list raw_input unichr callable format locals reduce unicode chr frozenset long reload vars classmethod getattr map repr xrange cmp globals max reversed zip compile hasattr memoryview round __import__ complex hash min set  delattr help next setattr  dict hex object slice  dir id oct sorted'
            return "class def lambda None True False " + special_func + built_in 
        else:
            return super().keywords(s)



#!/usr/bin/env python
import inspect
import textwrap
import time
import traceback
import logging

import numpy as np
from PyQt5 import QtWidgets, QtCore, QtGui
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QApplication, QMainWindow

from nspyre.config import load_config, load_meta_config
from nspyre.definitions import join_nspyre_path
from nspyre.misc import custom_decode
from nspyre.mongodb import Synched_Mongo_Database
from nspyre.spyrelet import load_spyrelet_class

__package__ = 'nspyre.gui'

from .image import ImageWidget
from .widgets.code_editor import Scintilla_Code_Editor, Monokai_Python_Lexer
from .widgets.plotting import LinePlotWidget, HeatmapPlotWidget
from .widgets.splitter_widget import Splitter, SplitterOrientation
from .widgets.views import Formatter, View

logger = logging.getLogger(__name__)


class CustomView:
    def __init__(self, code_editor, w1D, w2D, plot_layout, initial_df, initial_cache):
        self.is_updating = False
        self.editor = code_editor
        self.plot_layout = plot_layout
        self.plot_type = '1D'
        self.w = w1D
        self.ws = {'1D': w1D, '2D': w2D}
        self.valid_code = False
        self._source = "plot_type = '1D'\ndef plot(df, cache):\n    return {}"
        self.last_df = initial_df
        self.last_cache = initial_cache
        self.editor.request_run_signal.connect(self.analyze)

    def start_updating(self):
        self.analyze()
        self.w.clear()
        if self.valid_code:
            self.is_updating = True
    
    def stop_updating(self):
        self.is_updating = False

    def get_source(self):
        return self._source

    def analyze(self):
        self.valid_code = False
        source = self.editor.text()
        self._source = source
        try:
            exec(source)
        except:
            print("Error in executing the code")
            traceback.print_exc()
            return
        if not ('plot_type' in locals() and 'plot' in locals()):
            print("Missing either a 'plot_type' variable or a 'plot' function")
            return
        _plot_type, plot_fun = locals()['plot_type'], locals()['plot']
        if _plot_type in ['1D', '2D']:
            self.plot_type = _plot_type
            self.w = self.ws[_plot_type]
            self.plot_layout.setCurrentWidget(self.w)
        else:
            print("Invalid plot type!  Must be either '1D' or '2D'")
            return
        
        self.update_fun = plot_fun
        self.valid_code = True
        self.update(self.last_df, self.last_cache)

    def update(self, df, cache):
        self.last_df = df
        self.last_cache = cache
        if self.is_updating and self.valid_code:
            if self.plot_type == '1D':
                traces = self.update_fun(df, cache)
                for name, data in traces.items():
                    self.w.set(name, xs=data[0], ys=data[1])
            elif self.plot_type == '2D':
                im = np.array(self.update_fun(df, cache))
                self.w.set(im)


class BaseView:
    def __init__(self, view, w):
        self.view = view
        self.is_updating = False
        self.update_fun = view.update_fun
        self.init_formatter = view.get_formatter(self.w, 'init')
        self.update_formatter = view.get_formatter(self.w, 'update')
            
    def start_updating(self):
        self.w.clear()
        if self.init_formatter is not None:
            self.init_formatter(self.w)
        self.is_updating = True

    def stop_updating(self):
        self.is_updating = False

    def get_source(self):
        return textwrap.dedent(inspect.getsource(self.update_fun))


class LinePlotView(BaseView):
    def __init__(self, view, w=None):
        if view.type != '1D':
            raise TypeError('This class is for 1D plot only')
        if w is None:
            self.w = LinePlotWidget()
            self.w.setFont(QFont('Helvetica [Cronyx]', 12))
        else:
            self.w = w
        super().__init__(view, self.w)
            
    def update(self, df, cache):
        if self.is_updating:
            traces = self.update_fun(df, cache)
            for name, data in traces.items():
                if len(data) == 2:
                    self.w.set(name, xs=data[0], ys=data[1])
                elif len(data) == 3:
                    self.w.set(name, xs=data[0], ys=data[1], yerrs=data[2])
            if self.update_formatter is not None:
                self.update_formatter(self.w, df, cache)


class HeatmapPlotView(BaseView):
    def __init__(self, view, w=None):
        if view.type != '2D':
            raise TypeError('This class is for 2D plot only.')
        if w is None:
            self.w = HeatmapPlotWidget()
            self.w.setFont(QFont('Helvetica [Cronyx]', 12))
        else:
            self.w = w
        super().__init__(view, self.w)
            
    def update(self, df, cache):
        if self.is_updating:
            im = np.array(self.update_fun(df, cache))
            self.w.set(im)
            if self.update_formatter is not None:
                self.update_formatter(self.w, df, cache)


class Spyrelet_Views:
    def __init__(self, spyrelet_class):
        if type(spyrelet_class) is str:
            cfg_path = load_meta_config()
            cfg = load_config(cfg_path)
            spyrelet_class = load_spyrelet_class(spyrelet_class, cfg)

        self.views = {x:getattr(spyrelet_class, x) for x in dir(spyrelet_class) if type(getattr(spyrelet_class, x)) is View}
        formatters = [getattr(spyrelet_class, x) for x in dir(spyrelet_class) if type(getattr(spyrelet_class, x)) is Formatter]

        # Associate format_init and format_update functions
        for f in formatters:
            for view_name in f.view_list:
                if view_name in self.views:
                    self.views[view_name].add_formatter(f)

    def get_1D_views(self):
        return {x:self.views[x] for x in self.views if self.views[x].type=='1D'}

    def get_2D_views(self):
        return {x:self.views[x] for x in self.views if self.views[x].type=='2D'}

    # def init_plot(self, plots):
    #     """plots should be a dictionary with the view_names as keys and plot_objects as value"""
    #     for name, plot in plots.items():
    #         self.views[name].format_plot(plot, 'init')

    # def update_plot(self, view_name, df, plot_obj):
    #     view = self.views[view_name]
    #     view.update_fun(df)
    #     view.format_plot(plot_obj, 'update')


class ViewManagerWindow(QMainWindow):
    def __init__(self, mongodb_addr=None, db_name='Spyre_Live_Data', react_to_drop=False, db=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowTitle('NSpyre View Manager')
        self.setFont(QFont('Helvetica [Cronyx]'))
        if db is None:
            self.db = Synched_Mongo_Database(db_name, mongodb_addr=mongodb_addr)
        else:
            self.db = db
        # cleanup_register(mongodb_addr)

        self.views = dict()
        self.items = dict()
        self.last_updated = dict()
        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self.update_colors)
        self.timer.start(100)
        self.fade_rate = 2
        
        layout = QtWidgets.QHBoxLayout()
        # Build tree
        self.tree = QtWidgets.QTreeWidget()
        self.tree.setColumnCount(1)
        self.tree.setHeaderHidden(True)

        layout.addWidget(self.tree)

        # Build layout
        self.plot_layout = QtWidgets.QStackedLayout()
        plot_container = QtWidgets.QWidget()
        plot_container.setLayout(self.plot_layout)
        self.default_image = ImageWidget(str(join_nspyre_path('images/logo.png')))
        self.plot_layout.addWidget(self.default_image)
        
        self.common_lineplotwidget = LinePlotWidget()
        self.common_heatmapplotwidget = HeatmapPlotWidget()
        self.plot_layout.addWidget(self.common_lineplotwidget)
        self.plot_layout.addWidget(self.common_heatmapplotwidget)

        horizontal_splitter = Splitter(main_w=self.tree, side_w=plot_container,
                                       orientation=SplitterOrientation.vertical_right_button)
        horizontal_splitter.setSizes([1, 400])
        horizontal_splitter.setHandleWidth(10)

        self.code_editor = Scintilla_Code_Editor()
        self.code_editor.setLexer(Monokai_Python_Lexer(self.code_editor))

        vertical_splitter = Splitter(main_w=horizontal_splitter, side_w=self.code_editor,
                                     orientation=SplitterOrientation.horizontal_top_button)
        vertical_splitter.setSizes([200, 20])
        vertical_splitter.setHandleWidth(10)
        self.setCentralWidget(vertical_splitter)

        if 'Register' in self.db.dfs:
            names = list(self.db.get_df('Register').index)
            names.sort()
            for name in names:
                self.add_col(name, try_update=False)

        self.tree.currentItemChanged.connect(self.new_table_selection)

        # Connect db signals
        self.db.col_added.connect(self.add_col)
        self.db.updated_row.connect(self._update_plot)
        if react_to_drop:
            self.db.col_dropped.connect(self.del_col)
        self.show()

    def _update_plot(self, col_name, row):
        # add a half second delay between updates so system does not lock-up on frequently updated datasets:
        if (time.time() - self.last_updated[col_name]) > 0.5:
            if col_name != 'Register':
                self.last_updated[col_name] = time.time()
            self.update_plot(col_name, row)

    def update_plot(self, col_name, row):
        if col_name == 'Register':
            self.add_col(row.name)
            return
        for name, view in self.views[col_name].items():
            if col_name in self.db.get_df('Register').index and 'cache' in self.db.get_df('Register').loc[col_name] and not self.db.get_df('Register').loc[col_name]['cache'] is np.nan:
                cache = custom_decode(self.db.get_df('Register').loc[col_name]['cache'])
            else:
                cache = {}
            view.update(self.db.get_df(col_name), cache)
        
    def add_col(self, col_name, try_update=True):
        if col_name == 'Register':
            return
        sclass = self.db.get_df('Register').loc[col_name].name
        spyrelet_views = Spyrelet_Views(sclass)
        
        if col_name in self.views:
            return
        top = QtWidgets.QTreeWidgetItem(self.tree, [col_name])
        self.default_color = top.background(0)  # This is used to remember the default color when instantiating (to restore in update_color)

        self.items[col_name] = {'__top': top}
        self.views[col_name] = dict()
        self.last_updated[col_name] = time.time()

        for name, view in spyrelet_views.views.items():
            if view.type == '1D':
                self.views[col_name][name] = LinePlotView(view, self.common_lineplotwidget)
            elif view.type == '2D':
                self.views[col_name][name] = HeatmapPlotView(view, self.common_heatmapplotwidget)
            self.items[col_name][name] = QtWidgets.QTreeWidgetItem(0)  # QtGui.QStandardItem(name)
            self.items[col_name][name].setText(0, name)
            
            # self.plot_layout.addWidget(self.views[col_name][name].w)
            top.addChild(self.items[col_name][name])


        self.tree.insertTopLevelItem(0, top)
        try:
            self.update_plot(col_name, None)
        except:
            pass

    def make_new_custom_view(self, col_name):
        name_template = 'Custom {}'
        for i in range(100):
            name = name_template.format(i)
            if not name in self.views[col_name]:
                if col_name in self.db.get_df('Register').index and 'cache' in self.db.get_df('Register').loc[col_name] and not self.db.get_df('Register').loc[col_name]['cache'] is np.nan:
                    cache = custom_decode(self.db.get_df('Register').loc[col_name]['cache'])
                else:
                    cache = {}
                self.views[col_name][name] = CustomView(self.code_editor, self.common_lineplotwidget, self.common_heatmapplotwidget, self.plot_layout, self.db.get_df(col_name), cache)
                self.items[col_name][name] = QtWidgets.QTreeWidgetItem(0)
                self.items[col_name][name].setText(0, name)
                self.items[col_name]['__top'].addChild(self.items[col_name][name])
                break

    def del_col(self, col_name):
        if col_name == 'Register':
            return
        # for pname, view in self.views[col_name].items():
        #     self.plot_layout.removeWidget(view.w)
        #     view.w.deleteLater()
        for iname, item in self.items[col_name].items():
            self.tree.removeItemWidget(item,0)
        self.tree.takeTopLevelItem(self.tree.indexOfTopLevelItem(self.items[col_name]['__top']))
        self.views.pop(col_name)
        self.items.pop(col_name)
        self.last_updated.pop(col_name)
    
    def get_view_name(self, item):
            if item is None:
                return None, None
            txt = item.text(0)
            if not txt in self.views:
                spyrelet_name = item.parent().text(0)
                view_name = txt
                return spyrelet_name, view_name
            else:
                # Selected a top level item
                return txt, None

    def new_table_selection(self, cur, prev):
        spyrelet_name, view_name = self.get_view_name(cur)
        spyrelet_name_old, view_name_old = self.get_view_name(prev)
        if view_name_old is not None:
            self.views[spyrelet_name_old][view_name_old].stop_updating()
        if view_name is None:
            # Selected a top level item
            self.plot_layout.setCurrentWidget(self.default_image)
            return
        
        self.code_editor.setText(self.views[spyrelet_name][view_name].get_source())
        self.plot_layout.setCurrentWidget(self.views[spyrelet_name][view_name].w)
        self.views[spyrelet_name][view_name].start_updating()
        if spyrelet_name in self.db.dfs:
            self.update_plot(spyrelet_name, None)

    def update_colors(self):
        try:
            for col_name, last_time in self.last_updated.items():
                delta = (time.time()-last_time)/self.fade_rate
                if delta<1:
                    color = QtGui.QColor(0, 255, 0, int(max((1-delta)*255,0)))
                    self.items[col_name]['__top'].setBackground(0, QtGui.QBrush(color))
                else:
                    self.items[col_name]['__top'].setBackground(0, self.default_color)
        except:
            pass

    def keyPressEvent(self, ev):
        if type(ev)==QtGui.QKeyEvent:
            sname, vname = self.get_view_name(self.tree.currentItem())
            if ev.matches(QtGui.QKeySequence.Copy):
                # app = QtWidgets.QApplication.instance()
                # app.clipboard().setText(sname)
                self.make_new_custom_view(sname)
                ev.accept()
            if ev.matches(QtGui.QKeySequence.Save):
                # filename, other = QtWidgets.QFileDialog.getSaveFileName(None, 'Save this table to...', '', 'Comma Separated Value (*.csv)')
                # if filename:
                print(self.db.get_df(sname))
                    # df = self.df_table.model()._data
                    # df.to_csv(path_or_buf=filename, sep=',')
                ev.accept()


if __name__ == '__main__':
    import logging
    import sys
    from PyQt5.QtCore import Qt
    from nspyre.gui import NSpyreApp
    from nspyre.misc import nspyre_init_logger

    nspyre_init_logger(logging.INFO)

    logger.info('starting View Manager...')
    if hasattr(Qt, 'AA_EnableHighDpiScaling'):
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
        QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
    app = NSpyreApp([sys.argv])
    data_view_window = ViewManagerWindow(react_to_drop=False)
    sys.exit(app.exec())



"""
This file tests the functionality of the spyrelets

Author: Jacob Feder
Date: 11/13/2020
"""

###########################
# imports
###########################

# std
from pathlib import Path

# nspyre
from nspyre.spyrelet.spyrelet import load_all_spyrelets, unload_all_spyrelets, \
                                    unload_spyrelet, load_spyrelet, reload_all_spyrelets, \
                                    reload_spyrelet, SpyreletLoadError
from nspyre.inserv.gateway import InservGateway

###########################
# tests
###########################

class TestSpyrelets:
    def test_spyrelet_load(self, gateway):
        spyrelets = load_all_spyrelets(gateway)
        unload_all_spyrelets()
        # spyrelets = load_all_spyrelets(gateway)
        # unload_spyrelet('sweep')
        # try:
        #     spyrelets = load_all_spyrelets(gateway)
        #     raise Exception('this should have failed in the previous line')
        # except SpyreletLoadError:
        #     pass
        # load_spyrelet('sweep', gateway)
        # reload_all_spyrelets(gateway)
        # reload_spyrelet('sweep', gateway)



class View:
    """Some Views for plotting."""

    def __init__(self, f, plot_type):
        self.update_fun = f
        self.type = plot_type
        self.init_formatters = dict()
        self.update_formatters = dict()

    def add_formatter(self, formatter):
        if formatter.type == 'init':
            self.init_formatters[formatter.handled_class] = formatter.format_fun
        elif formatter.type == 'update':
            self.update_formatters[formatter.handled_class] = formatter.format_fun

    def get_formatter(self, plot, formatter_type):
        formatters = self.init_formatters if formatter_type == 'init' else self.update_formatters
        for c in formatters:
            if issubclass(type(plot), c):
                return formatters[c]


class Formatter:
    def __init__(self, f, formatter_type, handled_class, view_list):
        self.format_fun = f
        self.type = formatter_type
        self.handled_class = handled_class
        self.view_list = view_list


def Plot1D(fun):
    """Functions marked with this decorators should take a single argument (beyond self) which will be the dataframe representing the data
       The function marked must return a dict with the following format {'trace_name_1':[x1, y1], 'trace_name_2':[x2, y2], ...}"""
    return View(fun, '1D')


def Plot2D(fun):
    """Functions marked with this decorators should take a single argument (beyond self) which will be the dataframe representing the data
       The function marked must return a 2D ndarray to be plotted"""
    return View(fun, '2D')


def PlotFormatInit(class_type_handled, view_list):
    """Functions marked with this decorators will be called once when initializing the views.
       They should declare in the decorators argument what type of class they will handle.
       For example, if a function wants to do some extra formatting on any subclass of BasePlotWidget from nspyre.widgets.plotting then:
       
       @Plot1D()
       def my_plot(self, df):
           return {'trace1':[df.x.values, df.y1.values],  'trace2':[df.x.values, df.y2.values]}
       
       @PlotFormat(BasePlotWidget, ['my_plot'])
       def my_formating_function(self, plot):
           plot.xlabel = 'x axis (in a.u.)'

        There can only be one one PlotFormatInit function associated with a given plotting function
       """
    def PlotFormatInit_Decorator(fun):
        return Formatter(fun, 'init', class_type_handled, view_list)
    return PlotFormatInit_Decorator


def PlotFormatUpdate(class_type_handled, view_list):
    """Same idea as PlotFormatUpdate, but functions marked with this decorators will be called every time the plot is updated."""
    def PlotFormatUpdate_Decorator(fun):
        return Formatter(fun, 'update', class_type_handled, view_list)
    return PlotFormatUpdate_Decorator



from collections import OrderedDict, Iterable

from PyQt5 import QtWidgets, QtCore
from pyqtgraph import ComboBox

from nspyre.definitions import Q_

from .spinbox import SpinBox


def get_feat_widget(feat):
    if feat['values'] is not None:
        w = ComboBoxFeatWidget(feat['values'])
    elif isinstance(feat['value'], (int, float, Q_)) or not feat['units'] is None:
            opts = dict()
            if feat['units'] is not None:
                opts['unit'] = feat['units']
            if feat['limits'] is not None:
                opts['bounds'] = feat['limits']
            opts['dec'] = True
            opts['minStep'] = 1e-3
            opts['decimals'] = 10
            if isinstance(feat['value'], int):
                opts['int'] = True
                opts['minStep'] = 1
                opts['decimals'] = 10
            w = SpinBoxFeatWidget(opts)
    elif feat['value'] is None:
        w = LineEditFeatWidget(text = 'Unknown type')
        w.set_readonly(True)
        return w
    else:
        w = LineEditFeatWidget(text = feat['value']) 

    w.set_readonly(feat['readonly'])
    if (not feat['value'] is None) and (not feat['units'] is None):
        w.setter(Q_(feat['value'], feat['units']))
    else:
        w.setter(feat['value'])
    return w


class BaseFeatWidget(QtWidgets.QWidget):

    set_requested = QtCore.pyqtSignal(object) # This signal will be triggered externally when the display value needs to be changed (argument is the new value)
    go_clicked = QtCore.pyqtSignal() # This signal will be triggered when the "go button" is clicked
    read_clicked = QtCore.pyqtSignal()  # This signal will be triggered when the "read button" is clicked

    def __init__(self, parent=None):
        # The parent class will need to generate a self.val_w a getter and a setter
        if not (hasattr(self, 'val_w') and hasattr(self, 'setter') and hasattr(self, 'getter')):
            raise NotImplementedError("This class must be subclassed and implement a value widget (<val_w>) and getter/setter methods")

        super().__init__(parent=parent)

        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        # Take care of the value widget and an external which can be triggered externally to change the value displayed
        self.val_w.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        layout.addWidget(self.val_w)
        self.set_requested.connect(self.setter)
        
        # Go button 
        go_icon = self.style().standardIcon(QtWidgets.QStyle.SP_DialogOkButton)
        self.go_button = QtWidgets.QPushButton()
        self.go_button.setIcon(go_icon)
        self.go_button.clicked.connect(self.go_clicked)
        layout.addWidget(self.go_button)
        
        # Read button
        read_icon = self.style().standardIcon(QtWidgets.QStyle.SP_BrowserReload)
        self.read_button = QtWidgets.QPushButton()
        self.read_button.setIcon(read_icon)
        self.read_button.clicked.connect(self.read_clicked)
        layout.addWidget(self.read_button)
        return

    def set_readonly(self, readonly):
        if readonly:
            self.val_w.setEnabled(False)
            self.go_button.setEnabled(False)
        else:
            self.val_w.setEnabled(True)
            self.go_button.setEnabled(True)
        return


class SpinBoxFeatWidget(BaseFeatWidget):

    def __init__(self, opts=None, parent=None):
        if opts is None:
            opts = dict()
        self.val_w = SpinBox(**opts)
        super().__init__(parent=parent)
        return

    def setter(self, value):
        return self.val_w.setValue(value)

    def getter(self):
        return self.val_w.getValue()


class ComboBoxFeatWidget(BaseFeatWidget):

    def __init__(self, values, parent=None):
        self.values = values
        self.val_w = ComboBox()
        if isinstance(self.values, OrderedDict):
            for key, value in self.values.items():
                self.val_w.addItem(str(key), str(value))
        elif isinstance(self.values, dict):
            for key, value in sorted(self.values.items()):
                self.val_w.addItem(str(key), str(value))
        elif isinstance(self.values, set):
            for value in sorted(self.values):
                self.val_w.addItem(str(value), str(value))
        elif isinstance(self.values, Iterable):
            for value in self.values:
                self.val_w.addItem(str(value), str(value))
        else:
            raise TypeError('invalid type encountered while populating values')
        super().__init__(parent=parent)
        return

    def setter(self, value):
        index = self.val_w.findText(str(value))
        self.val_w.setCurrentIndex(index)
        return

    def getter(self):
        index = self.val_w.currentIndex()
        values = self.values
        if isinstance(values, OrderedDict):
            key = list(values.keys())[index]
        elif isinstance(values, dict):
            key = list(sorted(values.keys()))[index]
        elif isinstance(values, set):
            key = list(sorted(values))[index]
        else:
            key = values[index]
        return key


class LineEditFeatWidget(BaseFeatWidget):

    def __init__(self, parent=None, text=None):
        self.val_w = QtWidgets.QLineEdit()
        super().__init__(parent=parent)
        return

    def setter(self, value):
        self.val_w.setText(str(value))
        return

    def getter(self):
        value = self.val_w.text()
        return value
